from typing import Sized, final

from loguru import logger

from nova.core.motion_group import MotionGroup
from nova.api import models
from nova.gateway import ApiGateway


class Controller(Sized):
    def __init__(
        self, *, api_gateway: ApiGateway, cell: str, controller_instance: models.ControllerInstance
    ):
        self._api_gateway = api_gateway
        self._controller_api = api_gateway.controller_api
        self._motion_group_api = api_gateway.motion_group_api
        self._cell = cell
        self._name = controller_instance.controller
        self._activated_motion_group_ids: list[str] = []

    @property
    def name(self) -> str:
        return self._name

    @final
    async def __aenter__(self):
        motion_group_ids = await self.activated_motion_group_ids()
        self._activated_motion_group_ids = motion_group_ids
        logger.info(f"Found motion group {motion_group_ids}")
        return self

    @final
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for motion_group_id in self._activated_motion_group_ids:
            logger.info(f"Deactivating motion group {motion_group_id}")
            await self._motion_group_api.deactivate_motion_group(self._cell, motion_group_id)

    def __len__(self) -> int:
        return len(self._activated_motion_group_ids)

    def motion_group(self, motion_group_id: int = 0) -> MotionGroup:
        return MotionGroup(
            api_gateway=self._api_gateway,
            cell=self._cell,
            motion_group_id=f"{motion_group_id}@{self._name}",
        )

    def __getitem__(self, motion_group_id: int) -> MotionGroup:
        return self.motion_group(motion_group_id)

    async def activated_motion_group_ids(self) -> list[str]:
        activate_all_motion_groups_response = (
            await self._motion_group_api.activate_all_motion_groups(
                cell=self._cell, controller=self._name
            )
        )
        motion_groups = activate_all_motion_groups_response.instances
        return [mg.motion_group for mg in motion_groups]

    async def activated_motion_groups(self) -> list[MotionGroup]:
        motion_group_ids = await self.activated_motion_group_ids()
        return [self.motion_group(int(mg.split("@")[0])) for mg in motion_group_ids]



from abc import ABC, abstractmethod
from typing import Annotated, Any, AsyncGenerator, Callable, Literal, Union

import pydantic
import wandelbots_api_client as wb
from nova.motion_settings import MotionSettings
from nova.types.collision_scene import CollisionScene
from nova.types.pose import Pose


class Action(pydantic.BaseModel, ABC):
    @abstractmethod
    @pydantic.model_serializer
    def serialize_model(self):
        """Serialize the model to a dictionary"""


class WriteAction(Action):
    type: Literal["Write"] = "Write"
    key: str
    value: Any

    @pydantic.model_serializer
    def serialize_model(self):
        return wb.models.IOValue(io=self.key, boolean_value=self.value).model_dump()


class ReadAction(Action):
    type: Literal["Read"] = "Read"
    device_id: str
    key: str


class ReadPoseAction(Action):
    type: Literal["ReadPose"] = "ReadPose"
    device_id: str
    tcp: str | None = None


class ReadJointsAction(Action):
    type: Literal["ReadJoints"] = "ReadJoints"
    device_id: str


class CallAction(Action):
    type: Literal["Call"] = "Call"
    device_id: str
    key: str
    arguments: list


MS = MotionSettings
PoseOrVectorTuple = Union[
    Pose, tuple[float, float, float, float, float, float], tuple[float, float, float]
]


class Motion(Action, ABC):
    """Base model of a motion

    Args:
        type: the type of the motion
        settings: the settings of the motion

    """

    type: Literal["linear", "ptp", "circular", "joint_ptp", "spline"]
    target: Pose | tuple[float, ...]
    settings: MotionSettings = MotionSettings()

    @property
    def is_cartesian(self):
        return isinstance(self.target, Pose)


class UnresolvedMotion(Motion, ABC):
    @abstractmethod
    async def resolve(
        self,
        initial_joints: tuple[float, ...],
        collision_scene: CollisionScene | None,
        configuration: dict,
        moving_robot_identifier: str,
    ) -> tuple[list[Motion], tuple[float, ...]] | None:
        """Convert the motion to a list of motion primitives

        Args:
            initial_joints: Joint positions at start of motion
            collision_scene: The collision scene used to check collisions
            configuration: E.g. data of physical setup of robot system, cell, etc.
            moving_robot_identifier: The identifier of the robot that is moving in the scene

        Returns:
            Tuple of resolved motions and the joint position at the end of the motions. None, if the motion can't be resolved

        """


class Linear(Motion):
    """A linear motion

    Examples:
    >>> Linear(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=10))
    Linear(type='linear', target=Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6)), settings=MotionSettings(min_blending_velocity=None, position_zone_radius=None, joint_velocity_limits=None, joint_acceleration_limits=None, tcp_velocity_limit=10.0, tcp_acceleration_limit=None, tcp_orientation_velocity_limit=None, tcp_orientation_acceleration_limit=None))

    """

    type: Literal["linear"] = "linear"
    target: Pose

    def _to_api_model(self) -> wb.models.PathLine:
        """Serialize the model to the API model

        Examples:
        >>> Linear(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=10))._to_api_model()
        PathLine(target_pose=Pose2(position=[1, 2, 3], orientation=[4, 5, 6]), path_definition_name='PathLine')
        """
        return wb.models.PathLine(
            target_pose=wb.models.Pose2(**self.target.model_dump()), path_definition_name="PathLine"
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def lin(target: PoseOrVectorTuple, settings: MotionSettings = MotionSettings()) -> Linear:
    """Convenience function to create a linear motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        settings: the motion settings

    Returns: the linear motion

    Examples:
    >>> ms = MotionSettings(tcp_velocity_limit=10)
    >>> assert lin((1, 2, 3, 4, 5, 6), settings=ms) == Linear(target=Pose((1, 2, 3, 4, 5, 6)), settings=ms)
    >>> assert lin((1, 2, 3)) == lin((1, 2, 3, 0, 0, 0))

    """
    t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
    return Linear(target=Pose(t), settings=settings)


class PTP(Motion):
    """A point-to-point motion

    Examples:
    >>> PTP(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=30))
    PTP(type='ptp', target=Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6)), settings=MotionSettings(min_blending_velocity=None, position_zone_radius=None, joint_velocity_limits=None, joint_acceleration_limits=None, tcp_velocity_limit=30.0, tcp_acceleration_limit=None, tcp_orientation_velocity_limit=None, tcp_orientation_acceleration_limit=None))

    """

    type: Literal["ptp"] = "ptp"

    def _to_api_model(self) -> wb.models.PathCartesianPTP:
        """Serialize the model to the API model

        Examples:
        >>> PTP(target=Pose((1, 2, 3, 4, 5, 6)), settings=MotionSettings(tcp_velocity_limit=30))._to_api_model()
        PathCartesianPTP(target_pose=Pose2(position=[1, 2, 3], orientation=[4, 5, 6]), path_definition_name='PathCartesianPTP')
        """
        if not isinstance(self.target, Pose):
            raise ValueError("Target must be a Pose object")
        return wb.models.PathCartesianPTP(
            target_pose=wb.models.Pose2(**self.target.model_dump()),
            path_definition_name="PathCartesianPTP",
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def ptp(target: PoseOrVectorTuple, settings: MotionSettings = MotionSettings()) -> PTP:
    """Convenience function to create a point-to-point motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        settings: the motion settings

    Returns: the point-to-point motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert ptp((1, 2, 3, 4, 5, 6), settings=ms) == PTP(target=Pose((1, 2, 3, 4, 5, 6)), settings=ms)
    >>> assert ptp((1, 2, 3)) == ptp((1, 2, 3, 0, 0, 0))

    """
    if isinstance(target, Pose):
        target = target.to_tuple()

    t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
    return PTP(target=Pose(t), settings=settings)


class Circular(Motion):
    """A circular motion

    Args:
        intermediate: the intermediate pose

    """

    type: Literal["circular"] = "circular"
    intermediate: Pose

    def _to_api_model(self) -> wb.models.PathCircle:
        """Serialize the model to a dictionary

        Examples:
        >>> Circular(target=Pose((1, 2, 3, 4, 5, 6)), intermediate=Pose((10, 20, 30, 40, 50, 60)), settings=MotionSettings(tcp_velocity_limit=30))._to_api_model()
        PathCircle(via_pose=Pose2(position=[10, 20, 30], orientation=[40, 50, 60]), target_pose=Pose2(position=[1, 2, 3], orientation=[4, 5, 6]), path_definition_name='PathCircle')
        """
        if not isinstance(self.target, Pose):
            raise ValueError("Target must be a Pose object")
        if not isinstance(self.intermediate, Pose):
            raise ValueError("Intermediate must be a Pose object")
        return wb.models.PathCircle(
            target_pose=wb.models.Pose2(**self.target.model_dump()),
            via_pose=wb.models.Pose2(**self.intermediate.model_dump()),
            path_definition_name="PathCircle",
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def cir(
    target: PoseOrVectorTuple,
    intermediate: PoseOrVectorTuple,
    settings: MotionSettings = MotionSettings(),
) -> Circular:
    """Convenience function to create a circular motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        intermediate: the intermediate pose or vector. If the intermediate is a vector, the orientation is set to
            (0, 0, 0).
        settings: the motion settings

    Returns: the circular motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert cir((1, 2, 3, 4, 5, 6), (7, 8, 9, 10, 11, 12), settings=ms) == Circular(target=Pose((1, 2, 3, 4, 5, 6)), intermediate=Pose((7, 8, 9, 10, 11, 12)), settings=ms)
    >>> assert cir((1, 2, 3), (4, 5, 6)) == cir((1, 2, 3, 0, 0, 0), (4, 5, 6, 0, 0, 0))

    """
    t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
    i = (*intermediate, 0.0, 0.0, 0.0) if len(intermediate) == 3 else intermediate
    return Circular(target=Pose(t), intermediate=Pose(i), settings=settings)


class JointPTP(Motion):
    """A joint PTP motion

    Examples:
    >>> JointPTP(target=(1, 2, 3, 4, 5, 6), settings=MotionSettings(tcp_velocity_limit=30))
    JointPTP(type='joint_ptp', target=(1.0, 2.0, 3.0, 4.0, 5.0, 6.0), settings=MotionSettings(min_blending_velocity=None, position_zone_radius=None, joint_velocity_limits=None, joint_acceleration_limits=None, tcp_velocity_limit=30.0, tcp_acceleration_limit=None, tcp_orientation_velocity_limit=None, tcp_orientation_acceleration_limit=None))

    """

    type: Literal["joint_ptp"] = "joint_ptp"

    def _to_api_model(self) -> wb.models.PathJointPTP:
        """Serialize the model to the API model

        Examples:
        >>> JointPTP(target=(1, 2, 3, 4, 5, 6, 7), settings=MotionSettings(tcp_velocity_limit=30))._to_api_model()
        PathJointPTP(target_joint_position=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0], path_definition_name='PathJointPTP')
        """
        if not isinstance(self.target, tuple):
            raise ValueError("Target must be a tuple object")
        return wb.models.PathJointPTP(
            target_joint_position=list(self.target), path_definition_name="PathJointPTP"
        )

    @pydantic.model_serializer
    def serialize_model(self):
        return self._to_api_model().model_dump()


def jnt(target: tuple[float, ...], settings: MotionSettings = MotionSettings()) -> JointPTP:
    """Convenience function to create a joint PTP motion

    Args:
        target: the target joint configuration
        settings: the motion settings

    Returns: the joint PTP motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert jnt((1, 2, 3, 4, 5, 6), settings=ms) == JointPTP(target=(1, 2, 3, 4, 5, 6), settings=ms)

    """
    return JointPTP(target=target, settings=settings)


class Spline(Motion):
    """A spline motion

    Args:
        path_parameter: the path parameter between 0 and 1
        time: the time in seconds

    """

    type: Literal["spline"] = "spline"
    path_parameter: float = pydantic.Field(1, ge=0)
    time: float | None = pydantic.Field(default=None, ge=0)

    @pydantic.model_serializer
    def serialize_model(self):
        """Serialize the model to a dictionary

        Examples:
        >>> JointPTP(target=(1, 2, 3, 4, 5, 6, 7), settings=MotionSettings(tcp_velocity_limit=30)).model_dump()
        {'target_joint_position': [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0], 'path_definition_name': 'PathJointPTP'}
        """
        raise NotImplementedError("Spline motion is not yet implemented")


def spl(
    target: PoseOrVectorTuple,
    settings: MotionSettings = MotionSettings(),
    path_parameter: float = 1,
    time=None,
) -> Spline:
    """Convenience function to create a spline motion

    Args:
        target: the target pose or vector. If the target is a vector, the orientation is set to (0, 0, 0).
        settings: the motion settings
        path_parameter: the path parameter between 0 and 1
        time: the time in seconds

    Returns: the spline motion

    Examples:
    >>> ms = MotionSettings(tcp_acceleration_limit=10)
    >>> assert spl((1, 2, 3, 4, 5, 6), settings=ms) == Spline(target=Pose((1, 2, 3, 4, 5, 6)), settings=ms)
    >>> assert spl((1, 2, 3)) == spl((1, 2, 3, 0, 0, 0))

    """
    t = (*target, 0.0, 0.0, 0.0) if len(target) == 3 else target
    return Spline(target=Pose(t), settings=settings, path_parameter=path_parameter, time=time)


class ActionLocation(pydantic.BaseModel):
    """A container for an action at a specific path parameter"""

    path_parameter: float = 1.0
    action: WriteAction


# TODO: all actions should be allowed (Action)
ActionContainerItem = Motion | WriteAction


class CombinedActions(pydantic.BaseModel):
    """A trajectory of motions and actions"""

    # See: https://docs.pydantic.dev/latest/concepts/serialization/#serialize_as_any-runtime-setting
    items: tuple[
        Annotated[
            pydantic.SerializeAsAny[ActionContainerItem], pydantic.Field(discriminator="type")
        ],
        ...,
    ] = ()

    def __len__(self):
        return len(self.items)

    def __getitem__(self, item):
        return self.items[item]

    def __setattr__(self, key, value):
        if key == "items":
            raise TypeError("Cannot set items directly")
        super().__setattr__(key, value)

    def __iter__(self):
        return iter(self.items)

    def append(self, item: ActionContainerItem):
        super().__setattr__("items", self.items + (item,))

    def _generate_trajectory(self) -> tuple[list[Motion], list[ActionLocation]]:
        """Generate two lists: one of Motion objects and another of ActionContainer objects,
        where each ActionContainer wraps a non-Motion action with its path parameter.

        The path parameter is the index of the last Motion object in the list of Motion objects.
        S - M - M - A - A - M - M - A - M - M
        0 - 1 - 2 - 3 - 3 - 3 - 4 - 5 - 5 - 6

        Returns:
            tuple: A tuple containing:
                - list of Motion objects from self.items.
                - list of ActionContainer objects with indexed path parameters.
        """
        motions = []
        actions = []
        last_motion_index = 0

        for item in self.items:
            if isinstance(item, Motion):
                motions.append(item)
                last_motion_index += 1  # Increment the motion index for each new Motion
            else:
                # Assign the current value of last_motion_index as path_parameter for actions
                actions.append(ActionLocation(path_parameter=last_motion_index, action=item))

        return motions, actions

    @property
    def motions(self) -> list[Motion]:
        motions, _ = self._generate_trajectory()
        return motions

    @property
    def actions(self) -> list[ActionLocation]:
        _, actions = self._generate_trajectory()
        return actions

    @property
    def start(self) -> ActionContainerItem | None:
        return self.motions[0] if self.motions else None

    @property
    def end(self) -> ActionContainerItem | None:
        return self.motions[-1] if self.motions else None

    def poses(self) -> list[Pose]:
        """Returns the positions of all motions. If a motion is not a cartesian motion, the position is ignored

        Returns: the positions

        """
        motions, _ = self._generate_trajectory()
        return [
            Pose(position=motion.target.position, orientation=motion.target.orientation)
            for motion in motions
            if motion.is_cartesian and isinstance(motion.target, Pose)
        ]

    def positions(self):
        """Returns the positions of all motions. If a motion is not a cartesian motion, the position is ignored

        Returns: the positions

        """
        return [pose.position for pose in self.poses()]

    def orientations(self):
        """Returns the orientations of all motions. If a motion is not a cartesian motion, the orientation is ignored

        Returns: the orientations

        """
        return [pose.orientation for pose in self.poses()]

    def __add__(self, other: "CombinedActions") -> "CombinedActions":
        return CombinedActions(items=self.items + other.items)

    def to_motion_command(self) -> list[wb.models.MotionCommand]:
        motion_commands = []
        for motion in self.motions:
            path = wb.models.MotionCommandPath.from_dict(motion.model_dump())
            blending = (
                motion.settings.as_blending_setting()
                if motion.settings.has_blending_settings()
                else None
            )
            limits_override = (
                motion.settings.as_limits_settings()
                if motion.settings.has_limits_override()
                else None
            )
            motion_commands.append(
                wb.models.MotionCommand(
                    path=path, blending=blending, limits_override=limits_override
                )
            )
        return motion_commands

    def to_set_io(self) -> list[wb.models.SetIO]:
        return [
            wb.models.SetIO(
                io=wb.models.IOValue(**action.action.model_dump(exclude_unset=True)),
                location=action.path_parameter,
            )
            for action in self.actions
        ]


class MovementControllerContext(pydantic.BaseModel):
    combined_actions: CombinedActions
    motion_id: str


ExecuteTrajectoryRequestStream = AsyncGenerator[wb.models.ExecuteTrajectoryRequest, None]
ExecuteTrajectoryResponseStream = AsyncGenerator[wb.models.ExecuteTrajectoryResponse, None]
MovementControllerFunction = Callable[
    [ExecuteTrajectoryResponseStream], ExecuteTrajectoryRequestStream
]
MovementController = Callable[[MovementControllerContext], MovementControllerFunction]



import asyncio
import functools
import time
from typing import TypeVar

import wandelbots_api_client as wb
from decouple import config
from loguru import logger

T = TypeVar("T")

INTERNAL_CLUSTER_NOVA_API = "http://api-gateway.wandelbots.svc.cluster.local:8080"


def intercept(api_instance: T):
    class Interceptor:
        def __init__(self, instance: T):
            self._instance = instance

        def __getattr__(self, name):
            # Retrieve the original attribute
            original_attr = getattr(self._instance, name)

            # If it's not callable, return it as is
            if not callable(original_attr):
                return original_attr

            # Wrap async callables
            if asyncio.iscoroutinefunction(original_attr):

                @functools.wraps(original_attr)
                async def async_wrapper(*args, **kwargs):
                    start = time.time()
                    try:
                        return await original_attr(*args, **kwargs)
                    except Exception as e:
                        logger.error(f"API CALL: {name} failed with error: {e}")
                        logger.debug(f"API CALL FAILED: {name} with args={args}, kwargs={kwargs}")
                        raise e
                    finally:
                        duration = time.time() - start
                        logger.info(f"API CALL: {name} took {duration:.2f} seconds")
                        logger.debug(f"API CALL: {name} with args={args}, kwargs={kwargs}")

                return async_wrapper

            # Wrap sync callables
            @functools.wraps(original_attr)
            def sync_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    return original_attr(*args, **kwargs)
                except Exception as e:
                    logger.error(f"API CALL: {name} failed with error: {e}")
                    raise e
                finally:
                    duration = time.time() - start
                    logger.info(f"API CALL: {name} took {duration:.2f} seconds")
                    logger.debug(f"API CALL: {name} with args={args}, kwargs={kwargs}")

            return sync_wrapper

    return Interceptor(api_instance)


class ApiGateway:
    def __init__(
        self,
        *,
        host: str | None = None,
        username: str | None = None,
        password: str | None = None,
        access_token: str | None = None,
        version: str = "v1",
        verify_ssl: bool = True,
    ):
        if host is None:
            host = config("NOVA_API", default=INTERNAL_CLUSTER_NOVA_API)

        if username is None:
            username = config("NOVA_USERNAME", default=None)

        if password is None:
            password = config("NOVA_PASSWORD", default=None)

        if access_token is None:
            access_token = config("NOVA_ACCESS_TOKEN", default=None)

        # Access token has more prio than username and password if both are provided at the same time, set username and
        # password to None
        if access_token is not None:
            username = None
            password = None

        stripped_host = host.rstrip("/")
        api_client_config = wb.Configuration(
            host=f"{stripped_host}/api/{version}",
            username=username,
            password=password,
            access_token=access_token,
        )
        api_client_config.verify_ssl = verify_ssl

        self._api_client = wb.ApiClient(configuration=api_client_config)
        self._host = host

        # Use the intercept function to wrap each API client
        self.controller_api = intercept(wb.ControllerApi(api_client=self._api_client))
        self.motion_group_api = intercept(wb.MotionGroupApi(api_client=self._api_client))
        self.motion_api = intercept(wb.MotionApi(api_client=self._api_client))
        self.motion_group_infos_api = intercept(wb.MotionGroupInfosApi(api_client=self._api_client))
        self.store_collision_components_api = intercept(
            wb.StoreCollisionComponentsApi(api_client=self._api_client)
        )
        self.store_collision_scenes_api = intercept(
            wb.StoreCollisionScenesApi(api_client=self._api_client)
        )
        self.virtual_robot_api = intercept(wb.VirtualRobotApi(api_client=self._api_client))
        self.virtual_robot_behavior_api = intercept(
            wb.VirtualRobotBehaviorApi(api_client=self._api_client)
        )
        self.virtual_robot_mode_api = intercept(wb.VirtualRobotModeApi(api_client=self._api_client))
        self.virtual_robot_setup_api = intercept(
            wb.VirtualRobotSetupApi(api_client=self._api_client)
        )

    async def close(self):
        return await self._api_client.close()



import time
import httpx
from pydantic import BaseModel, Field, ValidationError


class DeviceCodeInfo(BaseModel):
    """
    Model to store device code information.

    Attributes:
        device_code (str): The device code.
        user_code (str): The user code.
        verification_uri (str): The verification URI.
        expires_in (int): The expiration time in seconds.
        interval (int): The interval time in seconds (default is 5).
    """

    device_code: str
    user_code: str
    verification_uri: str
    expires_in: int
    interval: int = Field(default=5)


class TokenInfo(BaseModel):
    """
    Model to store token information.

    Attributes:
        access_token (str): The access token.
        refresh_token (str, optional): The refresh token.
    """

    access_token: str
    refresh_token: str | None = None


class Auth0Parameters(BaseModel):
    """
    Model to store Auth0 parameters.

    Attributes:
        auth0_domain (str): The Auth0 domain.
        auth0_client_id (str): The Auth0 client ID.
        auth0_audience (str): The Auth0 audience.
    """

    auth0_domain: str
    auth0_client_id: str
    auth0_audience: str


class Auth0DeviceAuthorization:
    """
    Class to handle Auth0 device authorization.

    Methods:
        __init__(auth0_domain: str, auth0_client_id: str, auth0_audience: str):
            Initializes the Auth0DeviceAuthorization instance with the given parameters.
        request_device_code():
            Requests a device code from Auth0.
        display_user_instructions():
            Displays instructions for the user to authenticate.
        poll_token_endpoint():
            Polls the token endpoint to obtain an access token.
        refresh_access_token(refresh_token: str):
            Refreshes the access token using the refresh token.
    """

    def __init__(self, auth0_domain: str, auth0_client_id: str, auth0_audience: str):
        """
        Initializes the Auth0DeviceAuthorization instance with the given parameters.

        Args:
            auth0_domain (str): The Auth0 domain.
            auth0_client_id (str): The Auth0 client ID.
            auth0_audience (str): The Auth0 audience.

        Raises:
            ValueError: If the parameters are not set correctly.
        """
        try:
            self.params = Auth0Parameters(
                auth0_domain=auth0_domain,
                auth0_client_id=auth0_client_id,
                auth0_audience=auth0_audience,
            )
        except ValidationError as e:
            raise ValueError(f"Error: The following parameters are not set correctly: {e}")

        self.auth0_domain = auth0_domain
        self.auth0_client_id = auth0_client_id
        self.auth0_audience = auth0_audience
        self.headers = {"content-type": "application/x-www-form-urlencoded"}
        self.device_code_info: DeviceCodeInfo | None = None
        self.interval = 5

    def request_device_code(self):
        """
        Requests a device code from Auth0.

        Returns:
            DeviceCodeInfo: The device code information.

        Raises:
            Exception: If there is an error requesting the device code.
        """
        device_code_url = f"https://{self.auth0_domain}/oauth/device/code"
        data = {
            "client_id": self.auth0_client_id,
            "scope": "openid profile email offline_access",
            "audience": self.auth0_audience,
        }

        response = httpx.post(device_code_url, headers=self.headers, data=data)
        if response.status_code == 200:
            self.device_code_info = DeviceCodeInfo(**response.json())
            self.interval = self.device_code_info.interval
            return self.device_code_info
        else:
            raise Exception("Error requesting device code:", response.json())

    def display_user_instructions(self):
        """
        Displays instructions for the user to authenticate.

        Raises:
            Exception: If device code information is not available.
        """
        if self.device_code_info:
            verification_uri = self.device_code_info.verification_uri
            user_code = self.device_code_info.user_code
            print(
                f"Please visit {verification_uri} and enter the code {user_code} to authenticate."
            )
        else:
            raise Exception("Device code information is not available.")

    def poll_token_endpoint(self):
        """
        Polls the token endpoint to obtain an access token.

        Returns:
            str: The access token.

        Raises:
            Exception: If there is an error polling the token endpoint.
        """
        if not self.device_code_info:
            raise Exception("Device code information is not available.")

        token_url = f"https://{self.auth0_domain}/oauth/token"
        token_data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": self.device_code_info.device_code,
            "client_id": self.auth0_client_id,
        }

        while True:
            token_response = httpx.post(token_url, headers=self.headers, data=token_data)
            if token_response.status_code == 200:
                # If the response status is 200, it means the access token is successfully obtained.
                token_info = TokenInfo(**token_response.json())
                self.refresh_token = token_info.refresh_token
                print("Access Token received:", token_info.access_token)
                print("Refresh Token received:", self.refresh_token)
                return token_info.access_token
            elif token_response.status_code == 400:
                # If the response status is 400, check the error type.
                error = token_response.json().get("error")
                if error == "authorization_pending":
                    # If the error is 'authorization_pending', it means the user has not yet authorized.
                    print("Waiting for user confirmation...")
                    time.sleep(self.interval)
                elif error == "slow_down":
                    # If the error is 'slow_down', it means the server requests to slow down polling.
                    print("Server requests to slow down polling. Increasing wait time.")
                    self.interval += 5
                    time.sleep(self.interval)
            elif token_response.status_code == 403:
                # If the response status is 403, it means the request is forbidden, wait and retry.
                print("Waiting...")
                time.sleep(self.interval)
            else:
                # For other status codes, raise an exception with the error details.
                raise Exception("Error:", token_response.status_code, token_response.json())

    def refresh_access_token(self, refresh_token: str):
        """
        Refreshes the access token using the refresh token.

        Args:
            refresh_token (str): The refresh token.

        Returns:
            str: The new access token.

        Raises:
            Exception: If there is an error refreshing the access token.
        """
        if not refresh_token:
            raise Exception("Refresh token is not available.")

        token_url = f"https://{self.auth0_domain}/oauth/token"
        token_data = {
            "grant_type": "refresh_token",
            "client_id": self.auth0_client_id,
            "refresh_token": refresh_token,
        }

        response = httpx.post(token_url, headers=self.headers, data=token_data)
        if response.status_code == 200:
            token_info = TokenInfo(**response.json())
            print("New Access Token received:", token_info.access_token)
            return token_info.access_token
        else:
            raise Exception("Error refreshing access token:", response.json())



import pydantic

from nova.types.pose import Pose


class RobotState(pydantic.BaseModel):
    """Collection of information on the current state of the robot"""

    pose: Pose
    joints: tuple[float, ...] | None = None


class MotionState(pydantic.BaseModel):
    """Collection of information on the current state of the robot"""

    path_parameter: float
    state: RobotState



import json

import wandelbots_api_client as wb


class ControllerNotFoundException(Exception):
    def __init__(self, controller: str):
        super().__init__(f"Controller {controller} not found.")


class PlanTrajectoryFailed(Exception):
    def __init__(self, error: wb.models.PlanTrajectoryFailedResponse):
        super().__init__(f"Plan trajectory failed: {json.dumps(error.to_dict(), indent=2)}")


class InitMovementFailed(Exception):
    def __init__(self, error: wb.models.InitializeMovementResponseInitResponse):
        super().__init__(f"Initial movement failed: {json.dumps(error.to_dict(), indent=2)}")


class LoadPlanFailed(Exception):
    def __init__(self, error: wb.models.PlanSuccessfulResponse):
        super().__init__(f"Load plan failed: {json.dumps(error.to_dict(), indent=2)}")



# This is the stream of requests that is send to execute trajectory websocket
from typing import AsyncGenerator, Callable

import wandelbots_api_client as wb

from nova.types.collision_scene import CollisionScene
from nova.types.pose import Pose
from nova.types.vector3d import Vector3d

LoadPlanResponse = wb.models.PlanSuccessfulResponse
InitialMovementStream = AsyncGenerator[wb.models.StreamMoveResponse, None]
InitialMovementConsumer = Callable[[wb.models.StreamMoveResponse], None]


__all__ = [
    "Vector3d",
    "Pose",
    "CollisionScene",
    "LoadPlanResponse",
    "InitialMovementStream",
    "InitialMovementConsumer",
]



import wandelbots_api_client as wb
from loguru import logger

from nova.actions import (
    ExecuteTrajectoryRequestStream,
    ExecuteTrajectoryResponseStream,
    MovementControllerContext,
    MovementControllerFunction,
)
from nova.core.exceptions import InitMovementFailed


def move_forward(context: MovementControllerContext) -> MovementControllerFunction:
    async def movement_controller(
        response_stream: ExecuteTrajectoryResponseStream,
    ) -> ExecuteTrajectoryRequestStream:
        # The first request is to initialize the movement
        yield wb.models.InitializeMovementRequest(trajectory=context.motion_id, initial_location=0)  # type: ignore

        # then we get the response
        initialize_movement_response = await anext(response_stream)
        if isinstance(
            initialize_movement_response.actual_instance, wb.models.InitializeMovementResponse
        ):
            r1 = initialize_movement_response.actual_instance
            if not r1.init_response.succeeded:
                raise InitMovementFailed(r1.init_response)

        # The second request is to start the movement
        set_io_list = context.combined_actions.to_set_io()
        yield wb.models.StartMovementRequest(
            set_ios=set_io_list, start_on_io=None, pause_on_io=None
        )  # type: ignore

        # then we wait until the movement is finished
        async for execute_trajectory_response in response_stream:
            r2 = execute_trajectory_response.actual_instance
            # Terminate the generator
            if isinstance(r2, wb.models.Standstill):
                if r2.standstill.reason == wb.models.StandstillReason.REASON_MOTION_ENDED:
                    return

    return movement_controller


def speed_up(context: MovementControllerContext) -> MovementControllerFunction:
    async def movement_controller(
        response_stream: ExecuteTrajectoryResponseStream,
    ) -> ExecuteTrajectoryRequestStream:
        # The first request is to initialize the movement
        yield wb.models.InitializeMovementRequest(trajectory=context.motion_id, initial_location=0)  # type: ignore

        # then we get the response
        initialize_movement_response = await anext(response_stream)
        if isinstance(
            initialize_movement_response.actual_instance, wb.models.InitializeMovementResponse
        ):
            r1 = initialize_movement_response.actual_instance
            if not r1.init_response.succeeded:
                raise InitMovementFailed(r1.init_response)

        # The second request is to start the movement
        set_io_list = context.combined_actions.to_set_io()
        yield wb.models.StartMovementRequest(
            set_ios=set_io_list, start_on_io=None, pause_on_io=None
        )  # type: ignore

        counter = 0
        latest_speed = 10
        # then we wait until the movement is finished
        async for execute_trajectory_response in response_stream:
            counter += 1
            response = execute_trajectory_response.actual_instance
            # Terminate the generator
            if isinstance(response, wb.models.Standstill):
                if response.standstill.reason == wb.models.StandstillReason.REASON_MOTION_ENDED:
                    return

            if isinstance(response, wb.models.PlaybackSpeedResponse):
                playback_speed = response.playback_speed_response
                logger.info(f"Current playback speed: {playback_speed}")

            if counter % 10 == 0:
                yield wb.models.ExecuteTrajectoryRequest(
                    wb.models.PlaybackSpeedRequest(playback_speed_in_percent=latest_speed)
                )
                counter = 0
                latest_speed += 5
                if latest_speed > 100:
                    latest_speed = 100

    return movement_controller



import pydantic
import wandelbots_api_client as wb


class MotionSettings(pydantic.BaseModel):
    """
    Settings for an action. This is closely related to the `MotionCommand` in the API.
    See planTrajectory.motion_commands for more information.

    Motion settings are immutable; if you need to change a setting, create a copy and update the new object.

    Attributes:
        min_blending_velocity:
            A minimum velocity for blending, in percent. Cannot be used if `blending` is set.

        position_zone_radius:
            Defines the position zone radius.

        joint_velocity_limits:
            Maximum joint velocities

        joint_acceleration_limits:
            Maximum joint accelerations

        tcp_velocity_limit:
            Maximum TCP velocity

        tcp_acceleration_limit:
            Maximum TCP acceleration

        tcp_orientation_velocity_limit:
            Maximum TCP orientation velocity

        tcp_orientation_acceleration_limit:
            Maximum TCP orientation acceleration
    """

    min_blending_velocity: int | None = pydantic.Field(default=None)
    position_zone_radius: float | None = pydantic.Field(default=None)
    joint_velocity_limits: tuple[float, ...] | None = pydantic.Field(default=None)
    joint_acceleration_limits: tuple[float, ...] | None = pydantic.Field(default=None)
    tcp_velocity_limit: float | None = pydantic.Field(default=None)
    tcp_acceleration_limit: float | None = pydantic.Field(default=None)
    tcp_orientation_velocity_limit: float | None = pydantic.Field(default=None)
    tcp_orientation_acceleration_limit: float | None = pydantic.Field(default=None)

    class Config:
        frozen = True

    @pydantic.model_validator(mode="after")
    def validate_blending_settings(self) -> "MotionSettings":
        if self.min_blending_velocity and self.position_zone_radius:
            raise ValueError("Can't set both min_blending_velocity and blending")
        return self

    def has_blending_settings(self) -> bool:
        return any([self.min_blending_velocity, self.position_zone_radius])

    def has_limits_override(self) -> bool:
        return any(
            [
                self.tcp_velocity_limit,
                self.tcp_acceleration_limit,
                self.tcp_orientation_velocity_limit,
                self.tcp_orientation_acceleration_limit,
                self.joint_velocity_limits,
                self.joint_acceleration_limits,
            ]
        )

    def as_limits_settings(self) -> wb.models.LimitsOverride:
        return wb.models.LimitsOverride(
            joint_velocity_limits=wb.models.Joints(joints=self.joint_velocity_limits)  # type: ignore
            if self.joint_velocity_limits
            else None,
            joint_acceleration_limits=wb.models.Joints(joints=self.joint_acceleration_limits)  # type: ignore
            if self.joint_acceleration_limits
            else None,
            tcp_velocity_limit=self.tcp_velocity_limit,
            tcp_acceleration_limit=self.tcp_acceleration_limit,
            tcp_orientation_velocity_limit=self.tcp_orientation_velocity_limit,
            tcp_orientation_acceleration_limit=self.tcp_orientation_acceleration_limit,
        )

    def as_blending_setting(self) -> wb.models.MotionCommandBlending:
        if self.position_zone_radius:
            return wb.models.MotionCommandBlending(
                wb.models.BlendingPosition(
                    position_zone_radius=self.position_zone_radius, blending_name="BlendingPosition"
                )
            )
        return wb.models.MotionCommandBlending(
            wb.models.BlendingAuto(
                min_velocity_in_percent=self.min_blending_velocity, blending_name="BlendingAuto"
            )
        )



import wandelbots_api_client as wb
from loguru import logger

from nova.actions import Action, CombinedActions, MovementController, MovementControllerContext
from nova.core.exceptions import LoadPlanFailed, PlanTrajectoryFailed
from nova.core.movement_controller import move_forward
from nova.gateway import ApiGateway
from nova.types import InitialMovementConsumer, InitialMovementStream, LoadPlanResponse
from nova.types.pose import Pose

MAX_JOINT_VELOCITY_PREPARE_MOVE = 0.2
START_LOCATION_OF_MOTION = 0.0


class MotionGroup:
    def __init__(
        self, api_gateway: ApiGateway, cell: str, motion_group_id: str, is_activated: bool = False
    ):
        self._api_gateway = api_gateway
        self._motion_api_client = api_gateway.motion_api
        self._cell = cell
        self._motion_group_id = motion_group_id
        self._current_motion: str | None = None
        self._optimizer_setup: wb.models.OptimizerSetup | None = None
        self.is_activated = is_activated

    async def __aenter__(self):
        await self._api_gateway.motion_group_api.activate_motion_group(
            cell=self._cell, motion_group=self._motion_group_id
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._api_gateway.motion_group_api.deactivate_motion_group(
            cell=self._cell, motion_group=self._motion_group_id
        )

    @property
    def motion_group_id(self) -> str:
        return self._motion_group_id

    @property
    def current_motion(self) -> str | None:
        # if not self._current_motion:
        #    raise ValueError("No MotionId attached. There is no planned motion available.")
        return self._current_motion

    async def plan(self, actions: list[Action] | Action, tcp: str) -> wb.models.JointTrajectory:
        """Plan a trajectory for the given actions

        Args:
            actions (list[Action] | Action): The actions to be planned. Can be a single action or a list of actions.
                Only motion actions are considered for planning.
            tcp (str): The identifier of the tool center point (TCP)

        Returns:
            wb.models.JointTrajectory: The planned joint trajectory
        """
        if not isinstance(actions, list):
            actions = [actions]

        if len(actions) == 0:
            raise ValueError("No actions provided")

        motion_commands = CombinedActions(items=tuple(actions)).to_motion_command()  # type: ignore
        joints = await self.joints()
        robot_setup = await self._get_optimizer_setup(tcp=tcp)
        request = wb.models.PlanTrajectoryRequest(
            robot_setup=robot_setup,
            start_joint_position=list(joints),
            motion_commands=motion_commands,
        )

        plan_trajectory_response = await self._motion_api_client.plan_trajectory(
            cell=self._cell, plan_trajectory_request=request
        )
        if isinstance(
            plan_trajectory_response.response.actual_instance,
            wb.models.PlanTrajectoryFailedResponse,
        ):
            # TODO: handle partially executable path
            raise PlanTrajectoryFailed(plan_trajectory_response.response.actual_instance)
        return plan_trajectory_response.response.actual_instance

    async def execute(
        self,
        joint_trajectory: wb.models.JointTrajectory,
        tcp: str,
        actions: list[Action] | Action | None = None,
        # collision_scene: dts.CollisionScene | None,
        movement_controller: MovementController = move_forward,
        initial_movement_consumer: InitialMovementConsumer | None = None,
    ):
        """Execute a planned motion

        Args:
            joint_trajectory (wb.models.JointTrajectory): The planned joint trajectory
            tcp (str): The identifier of the tool center point (TCP)
            actions (list[Action] | Action | None): The actions to be executed. Defaults to None.
            movement_controller (MovementController): The movement controller to be used. Defaults to move_forward
            initial_movement_consumer (InitialMovementConsumer): A consumer for the initial movement
        """
        if actions is None:
            actions = []
        elif not isinstance(actions, list):
            actions = [actions]

        load_plan_response = await self._load_planned_motion(joint_trajectory, tcp)

        number_of_joints = await self._get_number_of_joints()
        joints_velocities = [MAX_JOINT_VELOCITY_PREPARE_MOVE] * number_of_joints
        movement_stream = await self.move_to_start_position(joints_velocities, load_plan_response)
        async for move_to_response in movement_stream:
            if initial_movement_consumer is not None:
                initial_movement_consumer(move_to_response)

        movement_controller_context = MovementControllerContext(
            combined_actions=CombinedActions(items=tuple(actions)),  # type: ignore
            motion_id=load_plan_response.motion,
        )
        _movement_controller = movement_controller(movement_controller_context)
        await self._api_gateway.motion_api.execute_trajectory(self._cell, _movement_controller)

    async def plan_and_execute(self, actions: list[Action] | Action, tcp: str):
        joint_trajectory = await self.plan(actions, tcp)
        await self.execute(joint_trajectory, tcp, actions=actions)

    async def _get_number_of_joints(self) -> int:
        spec = await self._api_gateway.motion_group_infos_api.get_motion_group_specification(
            cell=self._cell, motion_group=self.motion_group_id
        )
        return len(spec.mechanical_joint_limits)

    async def _get_optimizer_setup(self, tcp: str) -> wb.models.OptimizerSetup:
        if self._optimizer_setup is None:
            self._optimizer_setup = (
                await self._api_gateway.motion_group_infos_api.get_optimizer_configuration(
                    cell=self._cell, motion_group=self._motion_group_id, tcp=tcp
                )
            )
        return self._optimizer_setup

    async def _load_planned_motion(
        self, joint_trajectory: wb.models.JointTrajectory, tcp: str
    ) -> wb.models.PlanSuccessfulResponse:
        load_plan_response = await self._api_gateway.motion_api.load_planned_motion(
            cell=self._cell,
            planned_motion=wb.models.PlannedMotion(
                motion_group=self.motion_group_id,
                times=joint_trajectory.times,
                joint_positions=joint_trajectory.joint_positions,
                locations=joint_trajectory.locations,
                tcp=tcp,
            ),
        )

        if (
            load_plan_response.plan_failed_on_trajectory_response is not None
            or load_plan_response.plan_failed_on_trajectory_response is not None
        ):
            raise LoadPlanFailed(load_plan_response)

        return load_plan_response.plan_successful_response

    async def move_to_start_position(
        self, joint_velocities, load_plan_response: LoadPlanResponse
    ) -> InitialMovementStream:
        limit_override = wb.models.LimitsOverride()
        if joint_velocities is not None:
            limit_override.joint_velocity_limits = wb.models.Joints(joints=joint_velocities)

        move_to_trajectory_stream = (
            self._api_gateway.motion_api.stream_move_to_trajectory_via_joint_ptp(
                cell=self._cell, motion=load_plan_response.motion, location_on_trajectory=0
            )
        )
        return move_to_trajectory_stream

    async def stop(self):
        logger.debug(f"Stopping motion of {self}...")
        try:
            await self._motion_api_client.stop_execution(
                cell=self._cell, motion=self.current_motion
            )
            logger.debug(f"Motion {self.current_motion} stopped.")
        except ValueError as e:
            logger.debug(f"No motion to stop for {self}: {e}")

    async def get_state(self, tcp: str | None = None) -> wb.models.MotionGroupStateResponse:
        """Get the current state of the motion group

        Args:
            tcp (str): The identifier of the tool center point (TCP) to be used for tcp_pose in response. If not set,
                the flange pose is returned as tcp_pose.
        """
        response = await self._api_gateway.motion_group_infos_api.get_current_motion_group_state(
            cell=self._cell, motion_group=self.motion_group_id, tcp=tcp
        )
        return response

    async def joints(self) -> tuple:
        """Get the current joint positions"""
        state = await self.get_state()
        return tuple(state.state.joint_position.joints)

    async def tcp_pose(self, tcp: str | None = None) -> Pose:
        """Get the current TCP pose"""
        state = await self.get_state(tcp=tcp)
        return Pose(state.state.tcp_pose)

    async def tcps(self) -> list[wb.models.RobotTcp]:
        """Get the available tool center points (TCPs)"""
        response = await self._api_gateway.motion_group_infos_api.list_tcps(
            cell=self._cell, motion_group=self.motion_group_id
        )
        return response.tcps

    async def tcp_names(self) -> list[str]:
        """Get the names of the available tool center points (TCPs)"""
        return [tcp.id for tcp in await self.tcps()]



from typing import Any

import numpy as np
import pydantic
import wandelbots_api_client as wb


class Vector3d(pydantic.BaseModel):
    """A vector 3d class
    Examples:
    >>> Vector3d(x=10, y=20, z=30)
    Vector3d(x=10, y=20, z=30)
    """

    x: float | int
    y: float | int
    z: float | int

    def __neg__(self) -> "Vector3d":
        return type(self)(x=-self.x, y=-self.y, z=-self.z)

    def __add__(self, other: Any) -> "Vector3d":
        if isinstance(other, (float, int)):
            return type(self)(x=self.x + other, y=self.y + other, z=self.z + other)
        if isinstance(other, Vector3d):
            return type(self)(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)
        return NotImplemented

    def __radd__(self, other: Any) -> "Vector3d":
        return self.__add__(other)

    def __sub__(self, other: Any) -> "Vector3d":
        return self.__add__(-other)

    def __rsub__(self, other: Any) -> "Vector3d":
        if isinstance(other, (float, int)):
            return type(self)(x=other - self.x, y=other - self.y, z=other - self.z)
        if isinstance(other, Vector3d):
            return type(self)(x=other.x - self.x, y=other.y - self.y, z=other.z - self.z)
        return NotImplemented

    def __mul__(self, other: Any) -> "Vector3d":
        if isinstance(other, (float, int)):
            return type(self)(x=other * self.x, y=other * self.y, z=other * self.z)
        return NotImplemented

    def __rmul__(self, other: Any) -> "Vector3d":
        return self * other

    def __truediv__(self, other: Any) -> "Vector3d":
        if not isinstance(other, (float, int)):
            return NotImplemented
        return (1 / other) * self

    def __len__(self):
        return 3

    @classmethod
    def from_tuple(cls, value: tuple[float, float, float]) -> "Vector3d":
        """Create a new Vector3d from tuple

        Examples:
        >>> Vector3d.from_tuple((10, 20, 30))
        Vector3d(x=10, y=20, z=30)
        >>> Vector3d.from_tuple((10.0, 20.5, 30.2))
        Vector3d(x=10.0, y=20.5, z=30.2)
        """
        return cls(x=value[0], y=value[1], z=value[2])

    def to_tuple(self) -> tuple[float, float, float]:
        """Return the vector as a tuple

        Examples:
        >>> Vector3d(x=10, y=20, z=30).to_tuple()
        (10, 20, 30)
        >>> Vector3d(x=10.0, y=20.5, z=30.2).to_tuple()
        (10.0, 20.5, 30.2)
        """
        return self.x, self.y, self.z

    def to_quaternion(self):
        """Interpret the object as a rotation vector and convert it to a quaternion"""
        values = np.asarray(self)
        half_angle = np.linalg.norm(values) / 2
        return np.concatenate([np.cos(half_angle)[None], values * np.sinc(half_angle / np.pi) / 2])

    @pydantic.model_serializer
    def serialize_model(self) -> wb.models.Vector3d:
        """
        Examples:
        >>> Vector3d.from_tuple((1, 2, 3)).model_dump()
        {'x': 1, 'y': 2, 'z': 3}
        """
        return wb.models.Vector3d(x=self.x, y=self.y, z=self.z)



from decouple import config
import asyncio

from nova.core.controller import Controller
from nova.core.exceptions import ControllerNotFoundException
from nova.core.logging_setup import configure_logging
from nova.gateway import ApiGateway
from nova.api import models
from loguru import logger


class Nova:
    def __init__(
        self,
        *,
        host: str | None = None,
        username: str | None = None,
        password: str | None = None,
        access_token: str | None = None,
        version: str = "v1",
        verify_ssl: bool = True,
        log_level: str = "INFO",
    ):
        configure_logging(log_level)
        self._api_client = ApiGateway(
            host=host,
            username=username,
            password=password,
            access_token=access_token,
            version=version,
            verify_ssl=verify_ssl,
        )

    def cell(self, cell_id: str = config("CELL_NAME", default="cell")) -> "Cell":
        return Cell(self._api_client, cell_id)

    async def close(self):
        return await self._api_client.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class Cell:
    def __init__(self, api_gateway: ApiGateway, cell_id: str):
        self._api_gateway = api_gateway
        self._cell_id = cell_id

    async def _get_controller_instances(self) -> list[models.ControllerInstance]:
        response = await self._api_gateway.controller_api.list_controllers(cell=self._cell_id)
        return response.instances

    async def _get_controller_instance(self, name: str) -> models.ControllerInstance | None:
        """Get the controller instance

        Args:
            name (str): The name of the controller

        Return: models.ControllerInstance: The controller instance
        """
        controllers = await self._get_controller_instances()
        controller = next((c for c in controllers if c.controller == name), None)
        return controller

    async def _wait_for_controller_ready(self, name: str, timeout: int):
        """Waits for the controller to be available

        Args:
            name (str): The name of the controller
            timeout (int): The time to wait for the controller to be available in seconds
        """
        iteration = 0
        controller = await self._get_controller_instance(name)

        while iteration < timeout:
            if controller is not None:
                if (
                    controller.error_details == "Controller not initialized or disposed"
                    or controller.error_details == "Initializing controller connection."
                ):
                    await self._api_gateway.controller_api.get_current_robot_controller_state(
                        cell=self._cell_id, controller=controller.host
                    )
                elif controller.has_error:
                    # As long has an error its being initialized
                    logger.error(controller.error_details)
                else:
                    # Controller is ready
                    return

            logger.info(f"Waiting for {self._cell_id}/{name} controller availability")
            await asyncio.sleep(1)
            controller = await self._get_controller_instance(name)
            iteration += 1

        raise TimeoutError(f"Timeout waiting for {self._cell_id}/{name} controller availability")

    async def add_virtual_robot_controller(
        self,
        name: str,
        controller_type: models.VirtualControllerTypes,
        controller_manufacturer: models.Manufacturer,
        timeout: int = 25,
        # TODO: This is not optimal yet and will be automatically resolved in the new v2 API. The default configuration
        #   is only valid for UR.
        #   See: https://code.wabo.run/robotics/wbr/-/blob/develop/wbr/src/service/internal/config/HomeJoints.h
        position: str = "[1.170,-1.6585,1.4051,-1.5707,-1.5709,1.170,0]",
    ) -> "Controller":
        """Add a virtual robot controller to the cell

        Args:
            name (str): The name of the controller
            controller_type (models.VirtualControllerTypes): The type of the controller
            controller_manufacturer (models.Manufacturer): The manufacturer of the controller
            timeout (int): The time to wait for the controller to be available in seconds
            position (str): The initial position of the robot

        """
        await self._api_gateway.controller_api.add_robot_controller(
            cell=self._cell_id,
            robot_controller=models.RobotController(
                name=name,
                configuration=models.RobotControllerConfiguration(
                    models.VirtualController(
                        type=controller_type,
                        manufacturer=controller_manufacturer,
                        position=position,
                    )
                ),
            ),
            completion_timeout=timeout,
        )
        # Technically not needed because of the completion_timeout but it handles edge cases right now
        await self._wait_for_controller_ready(name, timeout)
        controller_instance = await self._get_controller_instance(name)
        if controller_instance is None:
            raise ControllerNotFoundException(controller=name)

        return Controller(
            api_gateway=self._api_gateway,
            cell=self._cell_id,
            controller_instance=controller_instance,
        )

    async def ensure_virtual_robot_controller(
        self,
        name: str,
        controller_type: models.VirtualControllerTypes,
        controller_manufacturer: models.Manufacturer,
    ) -> "Controller":
        controller = await self._get_controller_instance(name)
        if controller:
            return Controller(
                api_gateway=self._api_gateway, cell=self._cell_id, controller_instance=controller
            )
        return await self.add_virtual_robot_controller(
            name, controller_type, controller_manufacturer
        )

    async def controllers(self) -> list["Controller"]:
        controllers = await self._get_controller_instances()
        return [
            Controller(api_gateway=self._api_gateway, cell=self._cell_id, controller_instance=c)
            for c in controllers
        ]

    async def controller(self, name: str) -> "Controller":
        controller = await self._get_controller_instance(name)

        if controller is None:
            raise ControllerNotFoundException(controller=name)

        return Controller(
            api_gateway=self._api_gateway, cell=self._cell_id, controller_instance=controller
        )

    async def delete_robot_controller(self, name: str, timeout: int = 25):
        await self._api_gateway.controller_api.delete_robot_controller(
            cell=self._cell_id, controller=name, completion_timeout=timeout
        )



from nova import actions, api, types
from nova.core.controller import Controller
from nova.core.motion_group import MotionGroup
from nova.core.movement_controller import speed_up as speed_up_movement_controller
from nova.motion_settings import MotionSettings
from nova.core.nova import Cell, Nova

__all__ = [
    "Nova",
    "Cell",
    "MotionGroup",
    "Controller",
    "speed_up_movement_controller",
    "api",
    "types",
    "actions",
    "MotionSettings",
]



# Provide autogenerated NOVA API client
from wandelbots_api_client import *  # noqa
from wandelbots_api_client import api, api_client, configuration, exceptions, models

__all__ = [
    "models",
    "api",
    "api_client",
    "configuration",
    "exceptions",
    "ApiResponse",  # noqa: F405
    "ApiClient",  # noqa: F405
    "Configuration",  # noqa: F405
    "OpenApiException",  # noqa: F405
    "ApiTypeError",  # noqa: F405
    "ApiValueError",  # noqa: F405
    "ApiKeyError",  # noqa: F405
    "ApiAttributeError",  # noqa: F405
    "ApiException",  # noqa: F405
]



import sys

from loguru import logger


def configure_logging(log_level: str = "INFO"):
    logger.remove()
    logger.add(
        sys.stderr,
        level=log_level,
        format="<green>{time}</green> | <level>{level}</level> | <level>{message}</level>",
    )

    return logger



from typing import Iterable, Sized

import numpy as np
import pydantic
import wandelbots_api_client as wb
from scipy.spatial.transform import Rotation

from nova.types.vector3d import Vector3d


def _parse_args(*args):
    if len(args) == 1 and (
        isinstance(args[0], wb.models.Pose) or isinstance(args[0], wb.models.TcpPose)
    ):
        pos = args[0].position
        ori = args[0].orientation
        return {
            "position": Vector3d(x=pos.x, y=pos.y, z=pos.z),
            "orientation": Vector3d(x=ori.x, y=ori.y, z=ori.z),
        }
    if len(args) == 1 and isinstance(args[0], wb.models.Pose2):
        x1, y1, z1 = args[0].position
        x2, y2, z2 = args[0].orientation
        return {"position": Vector3d(x=x1, y=y1, z=z1), "orientation": Vector3d(x=x2, y=y2, z=z2)}
    if len(args) == 1 and isinstance(args[0], tuple):
        args = args[0]
    if len(args) == 6:
        x1, y1, z1, x2, y2, z2 = args
        return {"position": Vector3d(x=x1, y=y1, z=z1), "orientation": Vector3d(x=x2, y=y2, z=z2)}
    elif len(args) == 3:
        x1, y1, z1 = args
        return {"position": Vector3d(x=x1, y=y1, z=z1), "orientation": Vector3d(x=0, y=0, z=0)}
    else:
        raise ValueError("Invalid number of arguments for Pose")


class Pose(pydantic.BaseModel, Sized):
    """A pose (position and orientation)

    Example:
    >>> Pose(position=Vector3d(x=10, y=20, z=30), orientation=Vector3d(x=1, y=2, z=3))
    Pose(position=Vector3d(x=10, y=20, z=30), orientation=Vector3d(x=1, y=2, z=3))
    """

    position: Vector3d
    orientation: Vector3d

    def __init__(self, *args, **kwargs):
        """Parse a tuple into a dict

        Examples:
        >>> Pose((1, 2, 3, 4, 5, 6))
        Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6))
        >>> Pose((1, 2, 3))
        Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=0, y=0, z=0))
        >>> Pose(wb.models.Pose(position=wb.models.Vector3d(x=1, y=2, z=3), orientation=wb.models.Vector3d(x=4, y=5, z=6)))
        Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6))
        >>> Pose(wb.models.TcpPose(position=wb.models.Vector3d(x=1, y=2, z=3), orientation=wb.models.Vector3d(x=4, y=5, z=6), coordinate_system=None, tcp='Flange'))
        Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6))
        >>> Pose(wb.models.Pose2(position=[1, 2, 3], orientation=[4, 5, 6]))
        Pose(position=Vector3d(x=1, y=2, z=3), orientation=Vector3d(x=4, y=5, z=6))
        """
        if args:
            values = _parse_args(*args)
            super().__init__(**values)
        else:
            super().__init__(**kwargs)

    def __str__(self):
        return str(round(self).to_tuple())

    def __round__(self, n=None):
        if n is not None:
            raise NotImplementedError("Setting precision is not supported yet")
        pos_and_rot_vector = self.to_tuple()
        return Pose(
            tuple(
                [round(a, 1) for a in pos_and_rot_vector[:3]]
                + [round(a, 3) for a in pos_and_rot_vector[3:]]
            )
        )

    def __len__(self):
        return 6

    def to_tuple(self) -> tuple[float, float, float, float, float, float]:
        """Return the pose as a tuple

        Example:
        >>> Pose(position=Vector3d(x=10, y=20, z=30), orientation=Vector3d(x=1, y=2, z=3)).to_tuple()
        (10, 20, 30, 1, 2, 3)
        """
        return self.position.to_tuple() + self.orientation.to_tuple()

    def __getitem__(self, item):
        return self.to_tuple()[item]

    def __matmul__(self, other):
        """
        Pose concatenation combines two poses into a single pose that represents the cumulative effect of both
        transformations applied sequentially.

        Args:
            other: can be a Pose, or an iterable with 6 elements

        Returns:
            Pose: the result of the concatenation

        Examples:
        >>> Pose((1, 2, 3, 0, 0, 0)) @ Pose((1, 2, 3, 0, 0, 0))
        Pose(position=Vector3d(x=2.0, y=4.0, z=6.0), orientation=Vector3d(x=0.0, y=0.0, z=0.0))
        >>> Pose((1, 2, 3, 0, 0, 0)) @ [1, 2, 3, 0, 0, 0]
        Pose(position=Vector3d(x=2.0, y=4.0, z=6.0), orientation=Vector3d(x=0.0, y=0.0, z=0.0))
        >>> Pose((1, 2, 3, 0, 0, 0)) @ (1, 2, 3, 0, 0, 0)
        Pose(position=Vector3d(x=2.0, y=4.0, z=6.0), orientation=Vector3d(x=0.0, y=0.0, z=0.0))
        >>> def as_iterator(data):
        ...     for d in data:
        ...         yield d
        >>> Pose((1, 2, 3, 0, 0, 0)) @ as_iterator([1, 2, 3, 0, 0, 0])
        Pose(position=Vector3d(x=2.0, y=4.0, z=6.0), orientation=Vector3d(x=0.0, y=0.0, z=0.0))
        """
        if isinstance(other, Pose):
            transformed_matrix = np.dot(self.matrix, other.matrix)
            return self._matrix_to_pose(transformed_matrix)
        elif isinstance(other, Iterable):
            seq = tuple(other)
            return self.__matmul__(Pose(seq))

        else:
            raise ValueError(f"Cannot multiply Pose with {type(other)}")

    def transform(self, other) -> "Pose":
        return self @ other

    def _to_wb_pose(self) -> wb.models.Pose:
        """Convert to wandelbots_api_client Pose

        Examples:
        >>> Pose(position=Vector3d(x=10, y=20, z=30), orientation=Vector3d(x=1, y=2, z=3))._to_wb_pose()
        Pose(position=Vector3d(x=10, y=20, z=30), orientation=Vector3d(x=1, y=2, z=3), coordinate_system=None)
        """
        return wb.models.Pose(
            position=wb.models.Vector3d(**self.position.model_dump()),
            orientation=wb.models.Vector3d(**self.orientation.model_dump()),
        )

    def _to_wb_pose2(self) -> wb.models.Pose2:
        """Convert to wandelbots_api_client Pose

        Examples:
        >>> Pose(position=Vector3d(x=10, y=20, z=30), orientation=Vector3d(x=1, y=2, z=3))._to_wb_pose2()
        Pose2(position=[10, 20, 30], orientation=[1, 2, 3])
        """
        return wb.models.Pose2(
            position=list(self.position.to_tuple()), orientation=list(self.orientation.to_tuple())
        )

    @pydantic.model_serializer
    def serialize_model(self):
        """
        Examples:
        >>> Pose(position=Vector3d(x=10, y=20, z=30), orientation=Vector3d(x=1, y=2, z=3)).model_dump()
        {'position': [10, 20, 30], 'orientation': [1, 2, 3]}
        """
        return self._to_wb_pose2().model_dump()

    def _to_homogenous_transformation_matrix(self):
        """Converts the pose (position and rotation vector) to a 4x4 homogeneous transformation matrix."""
        rotation_vec = [self.orientation.x, self.orientation.y, self.orientation.z]
        rotation_matrix = Rotation.from_rotvec(rotation_vec).as_matrix()
        mat = np.eye(4)
        mat[:3, :3] = rotation_matrix
        mat[:3, 3] = [self.position.x, self.position.y, self.position.z]
        return mat

    def _matrix_to_pose(self, matrix: np.ndarray) -> "Pose":
        """Converts a homogeneous transformation matrix to a Pose."""
        rotation_matrix = matrix[:3, :3]
        position = matrix[:3, 3]
        rotation_vec = Rotation.from_matrix(rotation_matrix).as_rotvec()
        return Pose(
            (
                position[0],
                position[1],
                position[2],
                rotation_vec[0],
                rotation_vec[1],
                rotation_vec[2],
            )
        )

    def orientation_to_quaternion(self):
        values = np.asarray(self.orientation)
        half_angle = np.linalg.norm(values) / 2
        return np.concatenate([np.cos(half_angle)[None], values * np.sinc(half_angle / np.pi) / 2])

    @property
    def matrix(self) -> np.ndarray:
        """Returns the homogeneous transformation matrix."""
        return self._to_homogenous_transformation_matrix()



from typing import Literal

import pydantic
import wandelbots_api_client as wb

from nova.types.pose import Pose


class DhParameter(pydantic.BaseModel):
    """Denavit-Hartenberg parameters for a single joint.
    Applied in d/theta --> a/alpha order."""

    a: float = 0.0  # [mm]
    alpha: float = 0.0  # [rad]
    d: float = 0.0  # [mm]
    theta: float = 0.0  # [rad]
    reverse_rotation_direction: bool = False


class Sphere(pydantic.BaseModel):
    shape_type: Literal["sphere"] = (
        "sphere"  # this allows shape type identification from Union of shapes
    )
    radius: float


class Box(pydantic.BaseModel):
    shape_type: Literal["box"] = "box"
    size_x: float
    size_y: float
    size_z: float
    type: Literal["HOLLOW", "FULL"] = "FULL"


class Rectangle(pydantic.BaseModel):
    shape_type: Literal["rectangle"] = "rectangle"
    size_x: float
    size_y: float


class Plane(pydantic.BaseModel):
    shape_type: Literal["plane"] = "plane"


class Cylinder(pydantic.BaseModel):
    shape_type: Literal["cylinder"] = "cylinder"
    radius: float
    height: float


class Capsule(pydantic.BaseModel):
    shape_type: Literal["capsule"] = "capsule"
    radius: float
    cylinder_height: float


class RectangularCapsule(pydantic.BaseModel):
    shape_type: Literal["rectangular_capsule"] = "rectangular_capsule"
    radius: float
    sphere_center_distance_x: float
    sphere_center_distance_y: float


class ConvexHull(pydantic.BaseModel):
    shape_type: Literal["convex_hull"] = "convex_hull"
    vertices: list[wb.models.Vector3d]


class Collider(pydantic.BaseModel):
    shape: Sphere | Box | Rectangle | Plane | Cylinder | Capsule | RectangularCapsule | ConvexHull
    pose: Pose
    margin: float = 0.0  # [mm] This will increase the size of the shape in all dimensions. Can be used to keep a safe distance to the shape.


class CollisionRobotConfiguration(pydantic.BaseModel):
    """Configuration of a robot in the collision scene.

    Default link shapes are provided for all supported robots. Set use_default_link_shapes to True to apply them.

    link_attachements are additional shapes that are attached to the link reference frames.
    The reference frame of the link_attachements is obtained after applying all sets of DH-parameters
    from base to (and including) the specified index.
    Adjacent links in the chain are not checked for collision.

    The tool is treated like the last link in the chain and all its shapes are attached to the flange frame.
    """

    # TODO: https://docs.jacobirobotics.com/motion/robot.html#item-obstacle
    #     These guys have an item obstcale attached at the tcp pose --> seems convenient for handling tools

    use_default_link_shapes: bool = True  # if True, default shapes are used for all links
    # shapes to attach to link reference frames, additionally to default shapes
    link_attachements: dict[int, dict[str, Collider]] = {}
    tool: dict[str, Collider] = {}  # shapes that make up the tool


class CollisionRobot(pydantic.BaseModel):
    """A robot in the collision scene is a single chain of articulated rigid bodies with defined shapes.
    The end of the chain is called flange and all tool shapes and tool center points (TCP) are attached to the flange.
    Adjacent links in the chain are not checked for collision.
    """

    mounting: Pose  # pose of the robot base w.r.t. the scene root
    dh_parameters: list[DhParameter] = []
    joint_positions: list[float] = []  # [rad], size should match dh_parameters
    links: dict[int, dict[str, Collider]] = {}  # shapes attached to link reference frames
    tool: dict[str, Collider] = {}  # shapes that make up the tool


class CollisionScene(pydantic.BaseModel):
    """A Collision scene is a collection of static colliders and robots with a shared reference frame.
    Assumptions:
    - everything is defined w.r.t. scene root
    - robot identifier matches a robot identifier --> the robot is placed in the scene via its mounting
    - tool colliders are defined in the flange frames and link_attachement colliders are defined in the respective link frames
    """

    static_colliders: dict[str, Collider] = {}
    robots: dict[str, CollisionRobot] = {}

    def add_static_collider(self, identifier: str, collider: Collider):
        self.static_colliders[identifier] = collider


# initial_args (json) -> store (objects)

"""
scene = read(arg...)
add_static_collider(scene, "box", Box(size_x=1, size_y=1, size_z=1), Pose.from_tuple([0, 0, 0, 0, 0, 0]))
"""

__all__ = [
    "Box",
    "Capsule",
    "Collider",
    "CollisionRobot",
    "CollisionRobotConfiguration",
    "CollisionScene",
    "ConvexHull",
    "Cylinder",
    "Plane",
    "Rectangle",
    "RectangularCapsule",
    "Sphere",
]
